import dataclasses
import logging
import os
import pathlib
import tomllib
from typing import Dict, List, Union

import tomli_w

LOGGER = logging.getLogger(__name__)

@dataclasses.dataclass
class Pin:
	chip: str
	number: int
	name: str

@dataclasses.dataclass
class Relay:
	pin: Pin
	name: str
	state: bool

def _deserialize(data: List[Dict[str, str]]) -> List[Relay]:
	"Deserialize a list of relays from the config."
	return [Relay(
		name=relay["name"],
		pin=Pin(
			chip=relay["chip"],
			number=int(relay["pinnumber"]),
			name=relay["pinname"],
		)) for relay in data]

def _serialize(data: List[Relay]) -> List[Dict[str, str]]:
	"Serialize a list of relays to the config."
	return [{
		"chip": relay.pin.chip,
		"name": relay.name,
		"pinnumber": str(relay.pin.number),
		"pinname": relay.pin.name,
	} for relay in data]

class Relays:
	"Class for interacting with relays."
	def __init__(self, configpath: pathlib.Path):
		self.configpath = configpath
		try:
			with open(configpath, "rb") as f:
				content = tomllib.load(f)
				self.config = content
		except Exception as e:
			LOGGER.info("Unable to load config file: %s", e)
			self.config = {
				"relays": [],
			}
		self.relays = _deserialize(self.config["relays"])


	def chips_list(self) -> List[str]:
		raise NotImplementedError()

	def pins_list(self) -> List[Pin]:
		raise NotImplementedError()

	def relay_add(self, chip: str, pinnumber: int, name: str) -> None:
		self.config["relays"].append({
			"chip": chip,
			"name": name,
			"pinnumber": pinnumber,
		})
		with open(self.configpath, "wb") as f:
			tomli_w.dump({
				"relays": _serialize(self.config["relays"]),
			}, f)
		LOGGER.info("Wrote config file %s", self.configpath)

	def relays_list(self) -> List[Relay]:
		return self.relays
	
	def relay_on(self, name: str) -> bool:
		raise NotImplementedError()

	def relay_set(self, name: str, state: bool) -> None:
		raise NotImplementedError()

class RelaysFake(Relays):
	"Class for fake relays, useful for testing."
	def chips_list(self) -> List[str]:
		return ["chipA", "chipB"]

	def pins_list(self) -> List[Pin]:
		return [
			Pin(chip="chipA", number=1, name="CON2-P1"),
			Pin(chip="chipA", number=2, name="CON2-P2"),
			Pin(chip="chipA", number=3, name="CON2-P10"),
			Pin(chip="chipB", number=1, name="CON2-P3"),
			Pin(chip="chipB", number=2, name="CON2-P7"),
		]

	def relays_list(self) -> List[Relay]:
		raise NotImplementedError()
	
	def relay_on(self, name: str) -> bool:
		raise NotImplementedError()

	def relay_set(self, name: str, state: bool) -> None:
		raise NotImplementedError()

class RelaysReal(Relays):
	"Class for controlling real relays."
